# -*- coding: utf-8 -*-
from pydeps.render_context import RenderContext, Rankdir


def test_render_context():
    ctx = RenderContext()
    with ctx.graph():
        ctx.write_rule('a', 'b')
    assert 'a -> b' in ctx.text()
    assert 'b -> a' not in ctx.text()
    assert 'rankdir = TB' in ctx.text()


def test_render_context_reverse():
    # verify that rankdir is reversed in RenderBuffer, not RenderContext.
    ctx = RenderContext(reverse=True, rankdir=Rankdir.BOTTOM_TOP)
    with ctx.graph():
        ctx.write_rule('a', 'b')
    assert 'b -> a' in ctx.text()
    assert 'a -> b' not in ctx.text()
    assert 'rankdir = BT' in ctx.text()


def test_render_context_rankdir():
    ctx = RenderContext(rankdir=Rankdir.LEFT_RIGHT)
    with ctx.graph():
        pass
    text = ctx.text()
    assert 'rankdir = LR' in text



# -*- coding: utf-8 -*-
"""
Compatibility imports between py2/py3
"""
# pragma: nocover
try:
    from itertools import zip_longest                   # noqa
except ImportError:
    from itertools import izip_longest as zip_longest   # noqa

try:
    import configparser                                 # noqa
except ImportError:
    import ConfigParser as configparser                 # noqa



# -*- coding: utf-8 -*-
"""cli entrypoints.
"""
from __future__ import print_function
import json
import os
import sys

from pydeps.configs import Config
from . import py2depgraph, cli, dot, target
from .depgraph2dot import dep2dot  # , cycles2dot
import logging
from . import colors
log = logging.getLogger(__name__)


def _pydeps(trgt, **kw):
    # Pass args as a **kw dict since we need to pass it down to functions
    # called, but extract locally relevant parameters first to make the
    # code prettier (and more fault tolerant).
    # print("KW:", kw, '\n', os.getcwd())
    # print('abspath:', os.path.abspath(kw.get('deps_out')))
    # print('target', trgt.workdir)
    # print('target', trgt)
    colors.START_COLOR = kw.get('start_color')
    # show_cycles = kw.get('show_cycles')
    nodot = kw.get('no_dot')
    no_output = kw.get('no_output')
    output = kw.get('output')
    fmt = kw['format']
    show_svg = kw.get('show')
    deps_out = kw.get('deps_out')
    dot_out = kw.get('dot_out')
    # reverse = kw.get('reverse')
    if os.getcwd() != trgt.workdir:
        # the tests are calling _pydeps directoy
        os.chdir(trgt.workdir)

    dep_graph = py2depgraph.py2dep(trgt, **kw)

    if kw.get('show_deps'):
        cli.verbose("DEPS:")
        if deps_out:
            # make sure output files are written to sensible directories
            directory, _fname = os.path.split(deps_out)
            if not directory:
                deps_out = os.path.join(trgt.calling_dir, deps_out)
            with open(deps_out, 'w') as fp:
                fp.write(dep_graph.__json__())
        else:
            print(dep_graph.__json__())

    dotsrc = depgraph_to_dotsrc(trgt, dep_graph, **kw)

    if not nodot:
        if kw.get('show_dot'):
            cli.verbose("DOTSRC:")
            if dot_out:
                # make sure output files are written to sensible directories
                directory, _fname = os.path.split(dot_out)
                if not directory:
                    dot_out = os.path.join(trgt.calling_dir, dot_out)
                with open(dot_out, 'w') as fp:
                    fp.write(dotsrc)
            else:
                print(dotsrc)

        if not no_output:
            try:
                svg = dot.call_graphviz_dot(dotsrc, fmt)
            except OSError as cause:
                raise RuntimeError("While rendering {!r}: {}".format(output, cause))
            if fmt == 'svg':
                svg = svg.replace(b'</title>', b'</title><style>.edge>path:hover{stroke-width:8}</style>')

            try:
                with open(output, 'wb') as fp:
                    cli.verbose("Writing output to:", output)
                    fp.write(svg)
            except OSError as cause:
                raise RuntimeError("While writing {!r}: {}".format(output, cause))

            if show_svg:
                try:
                    dot.display_svg(kw, output)
                except OSError as cause:
                    helpful = ""
                    if cause.errno == 2:
                        helpful = " (can be caused by not finding the program to open this file)"
                    raise RuntimeError("While opening {!r}: {}{}".format(output, cause, helpful))


def depgraph_to_dotsrc(target, dep_graph, **kw):
    """Convert the dependency graph (DepGraph class) to dot source code.
    """
    # if kw.get('show_cycles'):
    #     dotsrc = cycles2dot(target, dep_graph, **kw)
    # el
    if not kw.get('no_dot'):
        dotsrc = dep2dot(target, dep_graph, **kw)
    else:
        dotsrc = None
    return dotsrc


def externals(trgt, **kwargs):
    """Return a list of direct external dependencies of ``pkgname``.
       Called for the ``pydeps --externals`` command.
    """
    kw = dict(
        T='svg', config=None, debug=False, display=None, exclude=[], exclude_exact=[],
        externals=True, format='svg', max_bacon=2**65, no_config=True, nodot=False,
        noise_level=2**65, no_show=True, output=None, pylib=True, pylib_all=True,
        show=False, show_cycles=False, show_deps=False, show_dot=False,
        show_raw_deps=False, verbose=0, include_missing=True, start_color=0
    )
    kw.update(kwargs)
    depgraph = py2depgraph.py2dep(trgt, **kw)
    pkgname = trgt.fname
    log.info("DEPGRAPH: %s", depgraph)
    pkgname = os.path.splitext(pkgname)[0]

    res = {}
    ext = set()

    for k, src in list(depgraph.sources.items()):
        if k.startswith('_'):
            continue
        if not k.startswith(pkgname):
            continue
        if src.imports:
            imps = [imp for imp in src.imports if not imp.startswith(pkgname)]
            if imps:
                for imp in imps:
                    ext.add(imp.split('.')[0])
                res[k] = imps
    # return res  # debug
    return list(sorted(ext))


def pydeps(**args):
    """Entry point for the ``pydeps`` command.

       This function should do all the initial parameter and environment
       munging before calling ``_pydeps`` (so that function has a clean
       execution path).
    """
    sys.setrecursionlimit(10000)
    _args = dict(iter(Config(**args))) if args else cli.parse_args(sys.argv[1:])
    _args['curdir'] = os.getcwd()
    inp = target.Target(_args['fname'])
    log.debug("Target: %r", inp)

    if _args.get('output'):
        _args['output'] = os.path.abspath(_args['output'])
    else:
        _args['output'] = os.path.join(
            inp.calling_dir,
            inp.modpath.replace('.', '_') + '.' + _args.get('format', 'svg')
        )

    with inp.chdir_work():
        # log.debug("Current directory: %s", os.getcwd())
        _args['fname'] = inp.fname
        _args['isdir'] = inp.is_dir

        if _args.get('externals'):
            del _args['fname']
            exts = externals(inp, **_args)
            print(json.dumps(exts, indent=4))
            # return exts  # so the tests can assert

        else:
            # this is the call you're looking for :-)
            try:
                return _pydeps(inp, **_args)
            except (OSError, RuntimeError) as cause:
                if log.isEnabledFor(logging.DEBUG):
                    # we only want to log the exception if we're in debug mode
                    log.exception("While running pydeps:")
                cli.error(str(cause))


def call_pydeps(file_or_dir, **kwargs):
    """Programatic entry point for pydeps.

       See :class:`pydeps.configs.Config` class for the available options.
    """
    sys.setrecursionlimit(10000)
    inp = target.Target(file_or_dir)
    log.debug("Target: %r", inp)
    config = Config(**kwargs)

    if config.output:
        config.output = os.path.abspath(config.output)
    else:
        config.output = os.path.join(
            inp.calling_dir,
            inp.modpath.replace('.', '_') + '.' + config.format
        )

    ctx = dict(iter(config))

    with inp.chdir_work():
        ctx['fname'] = inp.fname
        ctx['isdir'] = inp.is_dir
        if config.externals:
            del ctx['fname']
            return externals(inp, **ctx)

        return _pydeps(inp, **ctx)


if __name__ == '__main__':  # pragma: nocover
    pydeps()



# -*- coding: utf-8 -*-
import os
from pydeps.py2depgraph import py2dep
from pydeps.pydeps import _pydeps
from pydeps.target import Target

from tests.filemaker import create_files
from tests.simpledeps import empty, simpledeps


def test_file():
    files = """
        a.py: |
            import collections
    """
    with create_files(files) as workdir:
        assert simpledeps('a.py') == set()


def test_file_in_sub_directory():
    files = """
        foo:
            - a:
                - b.py: |
                    import c
                - c.py: ""

    """
    with create_files(files) as workdir:
        # os.chdir('foo')
        # t = Target('a/b.py')
        # with t.chdir_work():
        #     print("CURDIR:", os.getcwd(), os.listdir('.'))
        #     # deps = py2dep(t, **empty(log="DEBUG", no_dot=True, show_deps=True))
        #     deps = py2dep(t, **empty())
        #     print("DEPSx:", deps)
        #     rels = {f'{a.name} -> {b.name}' for a, b in deps}
        #     print("RELS:", rels)
        #     assert 'c -> b.py' in rels
        assert 'c -> b.py' in simpledeps('foo/a/b.py')  # , '-L DEBUG')


def test_file_in_directory():
    files = """
            - a:
                - b.py: |
                    import c
                - c.py: ""

    """
    with create_files(files) as workdir:
        assert 'c -> b.py' in simpledeps('a/b.py')


def test_file_pylib():
    files = """
        a.py: |
            import collections
    """
    with create_files(files) as workdir:
        assert 'collections -> a.py' in simpledeps('a.py', '--pylib')


def test_file_pyliball():
    files = """
        a.py: |
            import collections
    """
    with create_files(files) as workdir:
        assert 'collections -> a.py' in simpledeps('a.py', '--pylib --pylib-all')



import sys
import time
import struct
# from .mf.mf_next import *     # for debugging next version
import modulefinder
from modulefinder import (
    ModuleFinder as NativeModuleFinder
)
from importlib.util import MAGIC_NUMBER
import marshal
import dis
from . import mfimp

HAVE_ARGUMENT = dis.HAVE_ARGUMENT

# from stdlib's modulefinder
_PY_SOURCE = mfimp.PY_SOURCE
_PY_COMPILED = mfimp.PY_COMPILED
_PKG_DIRECTORY = mfimp.PKG_DIRECTORY

# monkey-patch broken modulefinder._find_module
# (https://github.com/python/cpython/issues/84530)
# in Python 3.8-3.10
if hasattr(modulefinder, '_find_module'):
    modulefinder._find_module = mfimp.find_module


def load_pyc(fp, mf=None):
    """Load a pyc file from a file object.
    """
    # adapted from https://github.com/nedbat/coveragepy/blob/master/lab/show_pyc.py#L21
    data = fp.read()
    pos = 0
    if data[:4] != MAGIC_NUMBER:
        raise ImportError("Bad magic number in .pyc file")

    pos += 4
    read_date_and_size = True
    if sys.version_info >= (3, 7):
        # 3.7 added a flags word
        flags = struct.unpack('<L', data[4:8])[0]
        pos += 4
        hash_based = bool(flags & 0x01)
        # check_source = bool(flags & 0x02)
        if hash_based:
            source_hash = data[pos:pos + 8]
            pos += 8
            read_date_and_size = False
            # print(f"hash {binascii.hexlify(source_hash)}")
            # print(f"check_source {check_source}")
    if read_date_and_size:
        moddate = data[pos:pos + 4]
        pos += 4
        # modtime = time.asctime(time.localtime(struct.unpack('<L', moddate)[0]))
        # print(f"moddate {binascii.hexlify(moddate)} ({modtime})")
        size = data[pos:pos + 4]
        pos += 4
        # size = fp.read(4)
        # print("pysize %s (%d)" % (binascii.hexlify(size), struct.unpack('<L', size)[0]))
    assert len(data) >= pos
    co = marshal.loads(data[pos:])
    return co


class ModuleFinder(NativeModuleFinder):
    def import_hook(self, name, caller=None, fromlist=None, level=-1):
        self.msg(3, "import_hook: name(%s) caller(%s) fromlist(%s) level(%s)" % (name, caller, fromlist, level))
        parent = self.determine_parent(caller, level=level)
        q, tail = self.find_head_package(parent, name)
        if q.shortname in ('__future__', 'future'):  # [pydeps] the future package causes recursion overflow
            return None
        m = self.load_tail(q, tail)
        if not fromlist:
            return q
        if m.__path__:
            self.ensure_fromlist(m, fromlist)
        return None

    def load_module(self, fqname, fp, pathname, file_info):
        # fqname = dotted module name we're loading
        suffix, mode, kind = file_info
        kstr = {
            _PKG_DIRECTORY: 'PKG_DIRECTORY',
            _PY_SOURCE: 'PY_SOURCE',
            _PY_COMPILED: 'PY_COMPILED',
        }.get(kind, 'unknown-kind')
        self.msgin(2, "load_module(%s) fqname=%s, fp=%s, pathname=%s" % (kstr, fqname, fp and "fp", pathname))

        if kind == _PKG_DIRECTORY:
            module = self.load_package(fqname, pathname)
            self.msgout(2, "load_module ->", module)
            return module

        if kind == _PY_SOURCE:
            txt = fp.read()
            txt += b'\n' if isinstance(txt, bytes) else '\n'
            co = compile(
                txt,
                pathname,
                'exec',            # compile code block
                dont_inherit=True  # [pydeps] don't inherit future statements from current environment
            )

        elif kind == _PY_COMPILED:
            # (see issue #191)
            try:
                co = load_pyc(fp, self)
            except ImportError:
                self.msgout(2, "raise ImportError: Bad magic number", pathname)
                raise ImportError("Bad magic number in %s" % pathname)

            # if fp.read(4) != MAGIC_NUMBER:
            #     self.msgout(2, "raise ImportError: Bad magic number", pathname)
            #     raise ImportError("Bad magic number in %s" % pathname)
            # fp.read(4)   # skip modification timestamp
            # co = marshal.load(fp)  # load marshalled code object.

            # adapted from https://github.com/nedbat/coveragepy/blob/master/lab/show_pyc.py#L21
            # if fp.read(4) != MAGIC_NUMBER:
            #     self.msgout(2, "raise ImportError: Bad magic number", pathname)
            #     raise ImportError("Bad magic number in %s" % pathname)

            # read_date_and_size = True
            # if sys.version_info >= (3, 7):
            #     # 3.7 added a flags word
            #     flags = struct.unpack('<L', fp.read(4))[0]
            #     hash_based = bool(flags & 0x01)
            #     check_source = bool(flags & 0x02)
            #     # print(f"flags 0x{flags:08x}")
            #     if hash_based:
            #         source_hash = fp.read(8)
            #         read_date_and_size = False
            #         # print(f"hash {binascii.hexlify(source_hash)}")
            #         # print(f"check_source {check_source}")
            # if read_date_and_size:
            #     moddate = fp.read(4)
            #     modtime = time.asctime(time.localtime(struct.unpack('<L', moddate)[0]))
            #     # print(f"moddate {binascii.hexlify(moddate)} ({modtime})")
            #     size = fp.read(4)
            #     # print("pysize %s (%d)" % (binascii.hexlify(size), struct.unpack('<L', size)[0]))
            # co = marshal.load(fp)

        else:
            co = None
        m = self.add_module(fqname)
        m.__file__ = pathname
        if co:
            if self.replace_paths:
                co = self.replace_paths_in_code(co)
            m.__code__ = co
            self.scan_code(co, m)
        self.msgout(2, "load_module ->", m)
        return m

    def scan_code(self, co, m):
        code = co.co_code   # noqa
        # if sys.version_info >= (3, 4):
        #     scanner = self.scan_opcodes
        # elif sys.version_info >= (2, 5):
        #     scanner = self.scan_opcodes_25
        # else:
        #     scanner = self.scan_opcodes_24
        scanner = self.scan_opcodes
        for what, args in scanner(co):
            if what == "store":
                name, = args
                m.globalnames[name] = 1
            elif what in ("import", "absolute_import"):
                fromlist, name = args
                have_star = 0
                if fromlist is not None:
                    if "*" in fromlist:
                        have_star = 1
                    fromlist = [f for f in fromlist if f != "*"]
                if what == "absolute_import":
                    level = 0
                else:
                    level = -1
                self._safe_import_hook(name, m, fromlist, level=level)
                if have_star:
                    # We've encountered an "import *". If it is a Python module,
                    # the code has already been parsed and we can suck out the
                    # global names.
                    mm = None
                    if m.__path__:
                        # At this point we don't know whether 'name' is a
                        # submodule of 'm' or a global module. Let's just try
                        # the full name first.
                        mm = self.modules.get(m.__name__ + "." + name)
                    if mm is None:
                        mm = self.modules.get(name)
                    if mm is not None:
                        m.globalnames.update(mm.globalnames)
                        m.starimports.update(mm.starimports)
                        if mm.__code__ is None:
                            m.starimports[name] = 1
                    else:
                        m.starimports[name] = 1
            elif what == "relative_import":
                level, fromlist, name = args
                if name:
                    self._safe_import_hook(name, m, fromlist, level=level)
                else:
                    parent = self.determine_parent(m, level=level)
                    # m is still the caller here... [bp]
                    self._safe_import_hook(parent.__name__, m, fromlist, level=0)
            else:
                # We don't expect anything else from the generator.
                raise RuntimeError(what)

        for c in co.co_consts:
            if isinstance(c, type(co)):
                self.scan_code(c, m)



import os
from pydeps.cli import parse_args
from pydeps.pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_cluster():
    files = """
        - bar_module:
            - __init__.py: ''
            - bar_a:
                - __init__.py: ''
                - a.py: |
                    from bar_module.bar_a import aa
                - aa.py: ''
            - bar_b:
                - __init__.py: ''
                - b.py: from bar_module.bar_a import aa
            - bar_c:
                - __init__.py: ''
                - c.py: |
                    from bar_module.bar_a import a
        - foo_module:
            - __init__.py: ''
            - foo_a:
                - __init__.py: ''
                - a.py: |
                    from . import aa
                - aa.py: ''
            - foo_c:
                - __init__.py: ''
                - c.py: |-
                    from foo_module.foo_a import aa
                    from bar_module.bar_c import c
    """
    with create_files(files) as workdir:
        args = parse_args(['foo_module', '--no-config', '--show-deps', '--cluster', '--max-cluster-size=100',
                           '--show-dot', '--dot-output', 'output.dot', '--no-show', '-LINFO', '-vv'])
        pydeps(**args)
        assert 'output.dot' in os.listdir(workdir)
        dot_output = open('output.dot').read()

        assert 'subgraph cluster_bar_module' in dot_output
        assert 'ltail="cluster_bar_module"' in dot_output




from pydeps.depgraph import Graph, GraphNode


class Source:
    def __init__(self, name):
        self.name = name
        self.imported_by = set()

def test_kosaraju():
    nodes = [GraphNode(Source(str(i))) for i in range(10)]
    edges = [
        (nodes[0], nodes[1]),
        (nodes[1], nodes[2]),
        (nodes[2], nodes[0]),
        (nodes[2], nodes[8]),
        (nodes[8], nodes[9]),
        (nodes[1], nodes[3]),
        (nodes[3], nodes[4]),
        (nodes[4], nodes[5]),
        (nodes[5], nodes[6]),
        (nodes[6], nodes[3]),
        (nodes[5], nodes[7]),
    ]
    graph = Graph(nodes, edges)
    # print(graph)
    # import pprint
    # pprint.pprint(graph)
    scc = graph.kosaraju()
    print(scc)
    assert scc == [
        {nodes[3], nodes[4], nodes[5], nodes[6]}, 
        {nodes[2], nodes[0], nodes[1]}, 
        {nodes[7]},
        {nodes[8]},
        {nodes[9]}
    ]



# -*- coding: utf-8 -*-
# from devtools import debug
from pydeps.cli import parse_args
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_pydeps_config_ini():
    files = """
        config.ini: |
            [pydeps]
            rankdir = BT
            exclude = 
                a
                c
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp', '--config=config.ini'])
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']


def test_pydeps_config_setupcfg_ini():
    files = """
        setup.cfg: |
            [pydeps]
            rankdir = BT
            exclude = 
                a
                c
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp'])
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']


def test_pydeps_config_json():
    files = """
        pydeps.json: |
            {
                "pydeps": {
                    "rankdir": "BT",
                    "exclude": ["a", "c"]
                }
            }
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp', '--config=pydeps.json'])
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']



def test_pydeps_config_pydeps_yaml():
    files = """
        pydeps.yml: |
            pydeps:
                rankdir: BT
                exclude: 
                    - a
                    - c
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp'])
        # debug(args)
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']


def test_pydeps_config_yaml():
    files = """
        config.yml: |
            pydeps:
                rankdir: BT
                exclude: 
                    - a
                    - c
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp', '--config=config.yml'])
        # debug(args)
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']


def test_pydeps_config_pyproject_toml():
    files = """
        pyproject.toml: |
            [tool.pydeps]
                rankdir = "BT"
                exclude = ["a", "c"]
            
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp'])
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']



def test_pydeps_config_toml():
    files = """
        config.toml: |
            [tool.pydeps]
                rankdir = "BT"
                exclude = ["a", "c"]
            
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        args = parse_args(['relimp', '--config=config.toml'])
        assert args['rankdir'] == "BT"
        assert args['exclude'] == ['a', 'c']



# -*- coding: utf-8 -*-
from collections import defaultdict
from io import StringIO
from contextlib import contextmanager
import textwrap
import enum


def to_unicode(s):
    try:
        return unicode(s)
    except NameError:
        return s


class Rankdir(enum.Enum):
    BOTTOM_TOP = 'BT'
    TOP_BOTTOM = 'TB'
    LEFT_RIGHT = 'LR'
    RIGHT_LEFT = 'RL'

    def reverse(self):
        return Rankdir(self.value[::-1])


class RenderContext(object):
    def __init__(self, out=None, reverse=False, rankdir=Rankdir.TOP_BOTTOM):
        self.out = out
        self.fp = StringIO()
        self.fillcolor = '#ffffff'
        self.fontcolor = '#000000'
        self.name = None
        self.concentrate = None
        self.compound = None
        self.width = 0.75
        self.reverse = reverse
        self.rankdir = rankdir

    @contextmanager
    def graph(self, **kw):
        """Set up a graphviz graph context.
        """
        self.name = kw.get('name', 'G')
        self.fillcolor = kw.get('fillcolor', '#ffffff')
        self.fontcolor = kw.get('fontcolor', '#000000')
        if kw.get('concentrate', True):
            self.concentrate = 'concentrate = true;'
        else:
            self.concentrate = ''
        self.compound = 'compound = true;' if kw.get('compound') else ''
        self.dedent("""
            digraph {self.name} {{
                {self.concentrate}
                {self.compound}
                rankdir = {self.rankdir.value};
                node [style=filled,fillcolor="{self.fillcolor}",fontcolor="{self.fontcolor}",fontname=Helvetica,fontsize=10];

        """.format(self=self))
        yield
        self.writeln('}')

    def text(self):
        """Get value of output stream (StringIO).
        """
        if self.out:
            self.out.close()  # pragma: nocover
        return self.fp.getvalue()

    def write_rule(self, a, b, **attrs):
        """a -> b [a1=x,a2=y];
        """
        if self.reverse:
            a, b = b, a
        with self.rule():
            self.write('%s -> %s' % (self._nodename(a), self._nodename(b)))
            # remove default values from output
            self._delattr(attrs, 'weight', 1)
            self._delattr(attrs, 'minlen', 1)
            self._delattr(attrs, 'len', 1)
            self.write_attributes(attrs)

    def write_node(self, a, **attrs):
        """a [a1=x,a2=y];
        """
        with self.rule():
            nodename = self._nodename(a)
            self.write(nodename)
            # remove default values from output
            self._delattr(attrs, 'label', nodename)
            self._delattr(attrs, 'fillcolor', self.fillcolor)
            self._delattr(attrs, 'fontcolor', self.fontcolor)
            self._delattr(attrs, 'width', self.width)
            self.write_attributes(attrs)

    # -- end of external/public interface --

    def write(self, txt):
        """Write ``txt`` to file and output stream (StringIO).
        """
        self.fp.write(to_unicode(txt))
        if self.out:
            self.out.write(txt)  # pragma: nocover

    def writeln(self, txt):
        """Write ``txt`` and add newline.
        """
        self.write(txt + '\n')

    def dedent(self, txt):
        """Write ``txt`` dedented.
        """
        self.write(textwrap.dedent(txt))

    def write_attributes(self, attrs):
        """Write comma separated attribute values (if exists).
        """
        if attrs:
            self.write(
                ' [' + ','.join('%s="%s"' % kv for kv in sorted(attrs.items())) + ']'
            )
        else:  # pragma: nocover
            pass

    def _nodename(self, x):
        """Return a valid node name.
        """
        return x.replace('.', '_')

    def _delattr(self, attr, key, value):
        if attr.get(key) == value:
            del attr[key]

    @contextmanager
    def rule(self):
        """Write indented rule.
        """
        self.write('    ')
        yield
        self.writeln(';')


class RenderBuffer(object):
    def __init__(self, target,
                 reverse=False,
                 rankdir=Rankdir.TOP_BOTTOM,
                 cluster=False,
                 min_cluster_size=0,
                 max_cluster_size=1,
                 keep_target_cluster=False,
                 collapse_target_cluster=False, 
                 remove_islands=False, **kw):
        self.target = target
        self.nodes = []
        self.rule_nodes = set()
        self.clusters = defaultdict(list)
        self.rules = {}
        self.reverse = reverse
        self.rankdir = Rankdir(rankdir)
        if self.reverse:
            self.rankdir = self.rankdir.reverse()
        self.cluster = cluster
        self.min_cluster_size = min_cluster_size
        self.max_cluster_size = max_cluster_size
        self.graph_attrs = {}
        self.keep_target_cluster = keep_target_cluster
        self.collapse_target_cluster = collapse_target_cluster
        self.remove_islands = remove_islands

    def _nodecolor(self, n):
        for node, attrs in self.nodes:
            if node == n:
                return attrs['fillcolor']
        return '#000000'

    def cluster_stats(self):
        maxnodes = max(len(v) for v in self.clusters.values())
        minnodes = min(len(v) for v in self.clusters.values())
        return minnodes, maxnodes

    def _remove_small_clusters(self):
        # remove clusters that are too small
        target_cluster = self._target_clusterid()

        _remove = []
        for clusterid, nodes in sorted(self.clusters.items()):
            if clusterid == target_cluster:
                # Target cluster must always be there, don't remove it even if it's small. We can get here
                # when --collapse-target-cluster flag is used.
                continue
            if len(nodes) < self.min_cluster_size:
                # print("REMOVING:CLUSTER:", clusterid, nodes)
                self.nodes += nodes
                _remove.append(clusterid)
        for _r in _remove:
            del self.clusters[_r]

    def _collapse_cluster(self, clusterid, nodes):
        """Add a single cluster node (with a label listing contents?)
           and change all rules to reference this node instead.
        """
        first_node, first_attrs = nodes[0]
        first_attrs['shape'] = 'folder'
        first_attrs['label'] = clusterid
        self.nodes.append((clusterid, first_attrs))

        for node, attrs in nodes:   # for each node in this cluster
            # check all rules for in/out relations
            rules = list(self.rules.items())
            self.rules = {}
            for (a, b), rule_attrs in rules:
                # orig = (a, b)
                if a == node:
                    a = clusterid
                if b == node:
                    b = clusterid
                # if orig != (a, b):
                #     print("CHANGED[{}|{}]: {} TO {}".format(clusterid, node, orig, (a, b)))
                self.rules[(a, b)] = rule_attrs

        del self.clusters[clusterid]

    def triage_clusters(self):
        target_cluster = self._target_clusterid()

        if not self.collapse_target_cluster and not self.keep_target_cluster:
            # don't put nodes from the target into a cluster
            self.nodes += self.clusters[target_cluster]
            del self.clusters[target_cluster]

        self._remove_small_clusters()

        # collapse target cluster if requested
        if self.collapse_target_cluster:
            self._collapse_cluster(target_cluster, self.clusters[target_cluster])

        # collapse clusters that are too big
        for clusterid, nodes in sorted(self.clusters.items()):
            if len(nodes) > self.max_cluster_size and clusterid != target_cluster:
                self._collapse_cluster(clusterid, nodes)

    def text(self):
        ctx = RenderContext(reverse=self.reverse, rankdir=self.rankdir)
        if self.cluster:
            self.triage_clusters()
            if self.clusters:   # are there any clusters left after triage?
                self.graph_attrs['compound'] = True
                self.graph_attrs['concentrate'] = False

        with ctx.graph(**self.graph_attrs):
            clusters = set()
            for clusterid, nodes in sorted(self.clusters.items()):
                clusters.add(clusterid)
                ctx.writeln('subgraph cluster_%s {' % clusterid)
                ctx.writeln('    label = %s;' % clusterid)
                for n, attrs in nodes:
                    ctx.write_node(n, **attrs)
                ctx.writeln('}')

            # non-clustered nodes
            for n, attrs in sorted(self.nodes, key=lambda x: x[0].lower()):
                ctx.write_node(n, **attrs)

            intercluster = set()
            for (a, b), attrs in sorted(self.rules.items(), key=lambda x: (x[0][0].lower(), x[0][1].lower())):  # sort by node name
                if a == b:
                    continue
                cida = self._clusterid(a)
                cidb = self._clusterid(b)
                if cida == cidb:
                    if self.reverse:
                        attrs['fillcolor'] = self._nodecolor(b)
                    else:
                        attrs['fillcolor'] = self._nodecolor(a)
                    ctx.write_rule(a, b, **attrs)
                elif cida in clusters and cidb in clusters:
                    if (cida, cidb) not in intercluster:
                        intercluster.add((cida, cidb))
                        if self.reverse:
                            attrs['lhead'] = 'cluster_' + cida
                            attrs['ltail'] = 'cluster_' + cidb
                            attrs['fillcolor'] = self._nodecolor(b)
                        else:
                            attrs['ltail'] = 'cluster_' + cida
                            attrs['lhead'] = 'cluster_' + cidb
                            attrs['fillcolor'] = self._nodecolor(a)
                        ctx.write_rule(a, b, **attrs)
                else:
                    if cida in clusters:
                        if self.reverse:
                            attrs['lhead'] = 'cluster_' + cida
                        else:
                            attrs['ltail'] = 'cluster_' + cida
                    if cidb in clusters:
                        if self.reverse:
                            attrs['ltail'] = 'cluster_' + cidb
                        else:
                            attrs['lhead'] = 'cluster_' + cidb
                    if self.reverse:
                        attrs['fillcolor'] = self._nodecolor(b)
                    else:
                        attrs['fillcolor'] = self._nodecolor(a)
                    ctx.write_rule(a, b, **attrs)
        return ctx.text()

    @contextmanager
    def graph(self, **kw):
        self.graph_attrs.update(kw)
        yield

    def _clusterid(self, n):
        return n.split('.')[0]

    def write_node(self, n, **attrs):
        if self.remove_islands and n not in self.rule_nodes:
            return
        clusterid = self._clusterid(n)
        if self.cluster:
            self.clusters[clusterid].append((n, attrs))
        else:
            self.nodes.append((n, attrs))

    def write_rule(self, a, b, **attrs):
        self.rule_nodes.add(a)
        self.rule_nodes.add(b)
        self.rules[(a, b)] = attrs

    def _target_clusterid(self):
        return self._clusterid(self.target.fname)



# -*- coding: utf-8 -*-
from pydeps.pydeps import pydeps
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_from_html5lib():
    files = """
        foo:
            - __init__.py
            - a.py: |
                from bar import py
        bar:
            - __init__.py
            - py.py: |
                barpy = 42
    """
    with create_files(files) as workdir:
        assert simpledeps('foo', '--show-deps -LINFO -vv') == {
            'bar -> foo.a',
            'bar.py -> foo.a'
        }


def test_multidot():
    files = """
        foo.bar.py: |
            from math import pi
    """
    with create_files(files) as workdir:
        assert simpledeps('foo.bar.py', '--show-deps --pylib -LINFO -vv') == {
            'math -> foo.bar.py',
        }



# -*- coding: utf-8 -*-
import os

from pydeps.colors import rgb2css, brightness, brightnessdiff, colordiff, name2rgb, foreground

red = (255, 0, 0)
green = (0, 255, 0)
yellow = (0, 255, 255)
blue = (0, 0, 255)
black = (0, 0, 0)
white = (255, 255, 255)


def test_rgb2css():
    assert rgb2css(red) == '#ff0000'
    assert rgb2css(green) == '#00ff00'
    assert rgb2css(yellow) == '#00ffff'
    assert rgb2css(blue) == '#0000ff'
    assert rgb2css(black) == '#000000'
    assert rgb2css(white) == '#ffffff'


def test_brightness():
    assert brightnessdiff(yellow, white) < brightnessdiff(yellow, black)


def test_colordiff():
    assert colordiff(blue, yellow) < colordiff(blue, red)


def test_foreground():
    assert foreground(black, red, green, yellow, blue, black, white) == white
    assert foreground(black, red, green, yellow, blue, black) == yellow
    assert foreground(black, red, green, blue, black) == green
    assert foreground(black, red, blue, black) == red
    assert foreground(black, blue, black) == blue


def test_name2rgb():
    def fg(name):
        return foreground(
            name2rgb(13),
            black, white)

    assert fg('hello') == fg('hello.world')



# -*- coding: utf-8 -*-
import ast

from pydeps.pydeps import pydeps, externals
from pydeps.target import Target
from tests.filemaker import create_files
from tests.simpledeps import simpledeps


def test_relative_imports(capsys):
    files = """
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('foo') == {
            'bar -> foo.a',
            'bar.b -> foo.a'
        }
        pydeps(fname='foo', externals=True)
        io = capsys.readouterr()
        assert ast.literal_eval(io.out) == ['bar']


def test_externals_with_nested_sourcefile():
    files = """
        foo:
            - __init__.py
            - a.py: |
                from bar import b
        bar:
            - __init__.py
            - b.py: |
                from baz import c
        baz:
            - __init__.py
            - c.py: print('hello world')
    """
    with create_files(files):
        assert simpledeps('foo') == {
            'bar -> foo.a',
            'bar.b -> foo.a',
        }
        ext = externals(
            Target('foo/a.py', use_calling_fname=True),
            pylib_all=False,
            pylib=False,
        )
        assert "bar" in ext



# -*- coding: utf-8 -*-
"""
command line interface (cli) code.
"""
# pylint: disable=line-too-long
from __future__ import print_function
import argparse

from pydeps.configs import Config
from .arguments import Arguments
# import json
# from .pycompat import configparser
import logging
import os
import sys
import subprocess
import textwrap
from . import __version__


def error(*args, **kwargs):  # pragma: nocover
    """Print an error message and exit.
    """
    kwargs['file'] = sys.stderr
    print("\n\tERROR:", *args, **kwargs)
    if args and args[0].startswith("[Errno 2] No such file or directory"):
        print("\t(Did you forget to include an __init__.py?)")
    sys.exit(1)


#: the (will become) verbose function
verbose = None


def _not_verbose(*args, **kwargs):  # pragma: nocover
    pass


verbose = _not_verbose


def _mkverbose(level):
    def _verbose(n, *args, **kwargs):
        if not isinstance(n, int):  # we're only interested in small integers
            # this allows the simpler usage cli.verbose(msg)
            args = (n,) + args
            n = 1
        if 0 < level <= n:
            print(*args, **kwargs)
    return _verbose


def _find_current_package():
    cwd = os.getcwd()
    while 'setup.py' not in os.listdir(cwd) and cwd != os.path.dirname(cwd):
        cwd = os.path.dirname(cwd)
    if 'setup.py' not in os.listdir(cwd):
        raise Exception("--find-package didn't find setup.py in current, or any parent, directory")
    os.chdir(cwd)
    package_name = subprocess.check_output("python setup.py --name", shell=True).decode('u8').strip()
    return package_name


def base_argparser(argv=()):
    """Initial parser that can set values for the rest of the parsing process.
    """
    global verbose
    verbose = _not_verbose

    _p = argparse.ArgumentParser(add_help=False)
    _p.add_argument('--debug', action='store_true', dest='debug', help="turn on all the show and verbose options (mainly for debugging pydeps itself)")
    _p.add_argument('--config', help="specify config file", metavar="FILE")
    _p.add_argument('--no-config', help="disable processing of config files", action='store_true')
    _p.add_argument('--version', action='store_true', help='print pydeps version')
    _p.add_argument('-L', '--log', help=textwrap.dedent('''
        set log-level to one of CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET.
    '''))
    _p.add_argument('--find-package', action='store_true', help="tries to automatically find the name of the current package.")
    _args, argv = _p.parse_known_args(argv)

    if _args.log:
        loglevels = "CRITICAL DEBUG ERROR FATAL INFO WARN"
        if _args.log not in loglevels:  # pragma: nocover
            error('legal values for the -L parameter are:', loglevels)
        loglevel = getattr(logging, _args.log)
    else:
        loglevel = None

    logging.basicConfig(
        level=loglevel,
        format='%(filename)s:%(lineno)d: %(levelname)s: %(message)s'
    )

    if _args.version:  # pragma: nocover
        print("pydeps v" + __version__)
        sys.exit(0)

    return _p, _args, argv  # return parsed and remaining args


def parse_args(argv=()):
    """Parse command line arguments, and return a dict.
    """
    _p, _args, argv = base_argparser(argv)
    find_package = _args.find_package
    config_files = []

    if not _args.no_config:  # process config files
        # extra config file specified with --config <fname> has highest precedence
        if _args.config:
            config_files.append(_args.config)

        # .pydeps file specified in current directory is next
        local_pydeps = os.path.join(os.getcwd(), '.pydeps')
        if os.path.exists(local_pydeps):
            config_files.append(local_pydeps)

        # pydeps.yml file specified in current directory is next
        local_pydeps = os.path.join(os.getcwd(), 'pydeps.yml')
        if os.path.exists(local_pydeps):
            config_files.append(local_pydeps)

        # is there a pyproject.toml file?
        pyproj = os.path.join(os.getcwd(), 'pyproject.toml')
        if os.path.exists(pyproj):
            config_files.append(pyproj)

        # is there a setup.cfg file?
        pyproj = os.path.join(os.getcwd(), 'setup.cfg')
        if os.path.exists(pyproj):
            config_files.append(pyproj)

        # finally the .pydeps file in the the user's homedir
        home = os.environ['USERPROFILE' if sys.platform == 'win32' else 'HOME']
        home_pydeps = os.path.join(home, '.pydeps')
        if os.path.exists(home_pydeps):
            config_files.append(home_pydeps)

    args = Arguments(config_files, debug=True, parents=[_p])

    if not find_package:
        args.add('fname', kind="FNAME:input", help='filename')
    else:
        args.add('--fname', kind="FNAME:input", help='filename')

    args.add('-v', '--verbose', default=0, dest='verbose', action='count', help="be more verbose (-vv, -vvv for more verbosity)")
    args.add('-o', default=None, kind="FNAME:output", dest='output', metavar="file", help="write output to 'file'")
    args.add('-T', default='svg', dest='format', help="output format (svg|png)")
    args.add('--display', kind="FNAME:exe", default=None, help="program to use to display the graph (png or svg file depending on the T parameter)", metavar="PROGRAM")
    args.add('--noshow', '--no-show', action='store_true', default=False, dest='no_show', help="don't call external program to display graph")
    args.add('--show-deps', action='store_true', help="show output of dependency analysis")
    args.add('--show-raw-deps', action='store_true', help="show output of dependency analysis before removing skips")
    args.add('--deps-output', dest='deps_out', default=None, kind="FNAME:output", help="write output of dependency analysis to 'file'")
    args.add('--show-dot', action='store_true', help="show output of dot conversion")
    args.add('--dot-output', dest='dot_out', default=None, kind="FNAME:output", help="write dot code to 'file'")
    args.add('--nodot', '--no-dot', action='store_true', default=False, dest='no_dot', help="skip dot conversion")
    args.add('--no-output', action='store_true', help="don't create .svg/.png file, implies --no-show (-t/-o will be ignored)")
    args.add('--show-cycles', action='store_true', help="show only import cycles")
    args.add('--debug-mf', default=0, type=int, metavar="INT", help="set the ModuleFinder.debug flag to this value")
    args.add('--noise-level', default=200, type=int, metavar="INT", help="exclude sources or sinks with degree greater than noise-level")
    args.add('--max-bacon', default=2, type=int, metavar="INT", help="exclude nodes that are more than n hops away (default=2, 0 -> infinite)")
    args.add('--max-module-depth', default=0, type=int, metavar="INT", help="coalesce deep modules to at most n levels")
    args.add('--pylib', action='store_true', help="include python std lib modules")
    args.add('--pylib-all', action='store_true', help="include python all std lib modules (incl. C modules)")
    args.add('--include-missing', action='store_true', help="include modules that are not installed (or can't be found on sys.path)")
    args.add('-x', '--exclude', default=[], nargs="+", metavar="PATTERN", help="input files to skip (e.g. `foo.*`), multiple file names can be provided")
    args.add('-xx', '--exclude-exact', default=[], nargs="+", metavar="MODULE", help="same as --exclude, except requires the full match. `-xx foo.bar` will exclude foo.bar, but not foo.bar.blob")
    args.add('--only', default=[], nargs="+", metavar="MODULE_PATH", help="only include modules that start with MODULE_PATH")
    args.add('--externals', action='store_true', help='create list of direct external dependencies')
    args.add('--reverse', action='store_true', help="draw arrows to (instead of from) imported modules")
    args.add('--rankdir', default='TB', type=str, choices=['TB', 'BT', 'LR', 'RL'], help="set the direction of the graph, legal values are TB (default, imported modules above importing modules), "
                                                                                         "BT (opposite direction of TB), LR (left-to-right), and RL (right-to-left)")
    args.add('--cluster', action='store_true', help="draw external dependencies as separate clusters")
    args.add('--min-cluster-size', default=0, type=int, metavar="INT", help="the minimum number of nodes a dependency must have before being clustered (default=0)")
    args.add('--max-cluster-size', default=0, type=int, metavar="INT", help="the maximum number of nodes a dependency can have before the cluster is collapsed to a single node (default=0)")
    args.add('--keep-target-cluster', action='store_true', help="draw target module as a cluster")
    args.add('--collapse-target-cluster', action='store_true', help="collapse target module (--keep-target-cluster will be ignored)")
    args.add('--rmprefix', default=[], nargs="+", metavar="PREFIX", help="remove PREFIX from the displayed name of the nodes")
    args.add('--start-color', default=0, type=int, metavar="INT", help="starting value for hue from 0 (red/default) to 360.")

    # args.write_default_config()
    _args = args.parse_args(argv)

    if _args.externals:
        return dict(vars(Config(
            externals=True, fname=_args.fname, max_bacon=10,
            include_missing=True, no_show=True
        )))

    if _args.no_output:
        _args.no_show = True
    _args.show = not _args.no_show
    if _args.no_dot and _args.show_cycles:
        error("Can't use --no=dot and --show-cycles together")  # pragma: nocover
    # if _args.show_cycles:
    #     _args.max_bacon = 0
    if _args.no_dot:
        _args.show_dot = False
    if _args.max_bacon == 0:
        _args.max_bacon = sys.maxsize
    if (
        _args.keep_target_cluster
        or _args.min_cluster_size > 0
        or _args.max_cluster_size > 0
        or _args.collapse_target_cluster
    ):
        _args.cluster = True
    if find_package:
        _args.fname = _find_current_package()

    _args.format = getattr(_args, 'format', 'svg')

    verbose = _mkverbose(max(_args.verbose, int(_args.debug)))
    verbose(2, _args, '\n')
    if _args.debug:  # pragma: nocover
        _args.verbose = 1
        _args.show = True
        _args.show_deps = True
        _args.show_dot = True

    return vars(_args)



    
from io import StringIO
import json
import warnings
import logging

log = logging.getLogger(__name__)

# from devtools import debug


HAVE_TOML = False
try:
    import tomllib as toml
    HAVE_TOML = True
except ImportError:
    try:
        import tomlkit as toml
        HAVE_TOML = True
    except ImportError:
        try:
            import toml
            HAVE_TOML = True
        except ImportError:
            pass


def is_string(v):
    return isinstance(v, str)
    

def boolval(v):
    if isinstance(v, bool):
        return v
    if isinstance(v, int):
        return bool(v)
    if is_string(v):
        v = v.lower()
        if v in {'j', 'y', 'ja', 'yes', '1', 'true'}:
            return True
        if v in {'n', 'nei', 'no', '0', 'false'}:
            return False
    raise ValueError("Don't know how to convert %r to bool" % v)


def listval(v):
    if is_string(v):
        return [x for x in v.split() if x.strip()]
    if isinstance(v, (list, tuple)):
        return v
    raise ValueError("Don't know how to convert %r to list" % v)


def identity(v):
    return v


typefns = {
    'BOOL': boolval,
    'INT': int,
    'LIST': listval,
    'STR': str,
}


def load_toml(filename):
    if not HAVE_TOML:
        return {}
    try:
        with open(filename) as fp:
            res = toml.loads(fp.read())
        return res['tool']['pydeps']
    except Exception as e:
        log.debug("Couldn't load toml file %s: %s", filename, e)
        return {}


def load_json(filename):
    try:
        with open(filename) as fp:
            res = json.loads(fp.read())
        return res['pydeps']
    except (json.JSONDecodeError, KeyError):
        return {}


def load_yaml(filename):
    try:
        import yaml
        from yaml import Loader
        with open(filename) as fp:
            res = yaml.load(fp.read(), Loader=Loader)
        return res['pydeps']
    except (yaml.YAMLError, KeyError, ImportError):
        return {}


def load_ini(filename):
    import configparser
    conf = configparser.ConfigParser()
    conf.read(filename)
    try:
        config = conf.items("pydeps")
        return dict(config)
    except (configparser.NoOptionError, configparser.NoSectionError):
        warnings.warn(' '.join("""
            Couldn't find a [pydeps] section in your config files
            %r -- or it was empty
        """.split()) % filename)
    return {}


def load_config(filename):
    if filename.endswith('.toml'):
        return load_toml(filename)
    if filename.endswith('.json'):
        return load_json(filename)
    if filename.endswith('.yaml'):
        return load_yaml(filename)
    raise ValueError("Unknown config file type: %s" % filename)


# this class is generated from pydeps.arguments.Arguments.write_default_config()
class Config(object):
    #: turn on all the show and verbose options (mainly for debugging pydeps
    #: itself)
    debug = False

    #: specify config file
    config = None

    #: disable processing of config files
    no_config = False

    #: print pydeps version
    version = False

    #:  set log-level to one of CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET.
    log = None

    #: tries to automatically find the name of the current package.
    find_package = False

    #: filename
    fname = None

    #: be more verbose (-vv, -vvv for more verbosity)
    verbose = 0

    #: write output to 'file'
    output = None

    #: output format (svg|png)
    format = 'svg'

    #: program to use to display the graph (png or svg file depending on the T
    #: parameter)
    display = None

    #: don't call external program to display graph
    no_show = False

    #: show output of dependency analysis
    show_deps = False

    #: show output of dependency analysis before removing skips
    show_raw_deps = False

    #: write output of dependency analysis to 'file'
    deps_out = None

    #: show output of dot conversion
    show_dot = False

    #: write dot code to 'file'
    dot_out = None

    #: skip dot conversion
    no_dot = False

    #: don't create .svg/.png file, implies --no-show (-t/-o will be ignored)
    no_output = False

    #: show only import cycles
    show_cycles = False

    #: set the ModuleFinder.debug flag to this value
    debug_mf = 0

    #: exclude sources or sinks with degree greater than noise-level
    noise_level = 200

    #: exclude nodes that are more than n hops away (default=2, 0 -> infinite)
    max_bacon = 2

    #: coalesce deep modules to at most n levels
    max_module_depth = 0

    #: include python std lib modules
    pylib = False

    #: include python all std lib modules (incl. C modules)
    pylib_all = False

    #: include modules that are not installed (or can't be found on sys.path)
    include_missing = False

    #: input files to skip (e.g. `foo.*`), multiple file names can be provided
    exclude = []

    #: same as --exclude, except requires the full match. `-xx foo.bar` will
    #: exclude foo.bar, but not foo.bar.blob
    exclude_exact = []

    #: only include modules that start with MODULE_PATH
    only = []

    #: create list of direct external dependencies
    externals = False

    #: draw arrows to (instead of from) imported modules
    reverse = False

    #: set the direction of the graph, legal values are TB (default, imported
    #: modules above importing modules), BT (opposite direction of TB),
    #: LR (left-to-right), and RL (right-to-left)
    rankdir = 'TB'

    #: draw external dependencies as separate clusters
    cluster = False

    #: the minimum number of nodes a dependency must have before being clustered
    #: (default=0)
    min_cluster_size = 0

    #: the maximum number of nodes a dependency can have before the cluster is
    #: collapsed to a single node (default=0)
    max_cluster_size = 0

    #: draw target module as a cluster
    keep_target_cluster = False

    #: collapse target module (--keep-target-cluster will be ignored)
    collapse_target_cluster = False

    #: remove PREFIX from the displayed name of the nodes
    rmprefix = []

    #: starting value for hue from 0 (red/default) to 360.
    start_color = 0

    def __init__(self, **kwargs):
        for key in dir(self.__class__):
            if not key.startswith('_'):
                val = getattr(self.__class__, key)
                if not callable(val):
                    setattr(self, key, val)
        self.__dict__.update(kwargs)

    def __eq__(self, other):
        res = True
        for (lkey, lval), (rkey, fval) in zip(iter(self), iter(other)):
            if lkey != rkey:
                print(f"key mismatch: {lkey} != {rkey}")
                res = False
            if lval != fval:
                print(f"val mismatch {lkey}: {lval} != {fval}")
                res = False
        return res

    def set_field(self, field, value):
        if field == 'fname':
            self.fname = identity(value)
        if field == 'verbose':
            self.verbose = int(value)
        if field == 'output':
            self.output = identity(value)
        if field == 'format':
            self.format = str(value)
        if field == 'display':
            self.display = identity(value)
        if field == 'no_show':
            self.no_show = boolval(value)
        if field == 'show_deps':
            self.show_deps = boolval(value)
        if field == 'show_raw_deps':
            self.show_raw_deps = boolval(value)
        if field == 'deps_out':
            self.deps_out = identity(value)
        if field == 'show_dot':
            self.show_dot = boolval(value)
        if field == 'dot_out':
            self.dot_out = identity(value)
        if field == 'no_dot':
            self.no_dot = boolval(value)
        if field == 'no_output':
            self.no_output = boolval(value)
        if field == 'show_cycles':
            self.show_cycles = boolval(value)
        if field == 'debug_mf':
            self.debug_mf = int(value)
        if field == 'noise_level':
            self.noise_level = int(value)
        if field == 'max_bacon':
            self.max_bacon = int(value)
        if field == 'max_module_depth':
            self.max_module_depth = int(value)
        if field == 'pylib':
            self.pylib = boolval(value)
        if field == 'pylib_all':
            self.pylib_all = boolval(value)
        if field == 'include_missing':
            self.include_missing = boolval(value)
        if field == 'exclude':
            self.exclude = listval(value)
        if field == 'exclude_exact':
            self.exclude_exact = listval(value)
        if field == 'only':
            self.only = listval(value)
        if field == 'externals':
            self.externals = boolval(value)
        if field == 'reverse':
            self.reverse = boolval(value)
        if field == 'rankdir':
            self.rankdir = str(value)
        if field == 'cluster':
            self.cluster = boolval(value)
        if field == 'min_cluster_size':
            self.min_cluster_size = int(value)
        if field == 'max_cluster_size':
            self.max_cluster_size = int(value)
        if field == 'keep_target_cluster':
            self.keep_target_cluster = boolval(value)
        if field == 'collapse_target_cluster':
            self.collapse_target_cluster = boolval(value)
        if field == 'rmprefix':
            self.rmprefix = listval(value)
        if field == 'start_color':
            self.start_color = int(value)

    def __iter__(self):
        return iter(self.__dict__.items())

    def __getattr__(self, name):
        return self.__dict__[name]

    def update(self, data):
        if set(data.keys()) - set(self.__dict__.keys()):
            warnings.warn("Unknown config keys: %s" % (set(data.keys()) - set(self.__dict__.keys())))
        for k, v in data.items():
            if k in self.__dict__:
                self.set_field(k, v)

    def write_ini(self):
        import configparser
        cfg = configparser.ConfigParser()
        cfg['pydeps'] = {k: str(v) for k, v in self}
        with StringIO() as fp:
            cfg.write(fp)
            return fp.getvalue()

    def write_toml(self):
        import tomlkit
        return tomlkit.dumps({
            'tools.pydeps': {k: v for k, v in self if v is not None}
        })

    def write_json(self):
        data = {'pydeps': self.__dict__}
        return json.dumps(data, indent=4)
    
    def write_yaml(self):
        import yaml
        data = {'pydeps': self.__dict__}
        return yaml.dump(data, default_flow_style=False)

    @classmethod
    def load(cls, fnames):
        """Load config from file.
        """
        conf = cls()

        for fname in fnames:
            ftype = filetype(fname)

            if ftype == 'toml':
                conf.update(load_toml(fname))
            if ftype == 'json':
                conf.update(load_json(fname))
            if ftype == 'yaml':
                conf.update(load_yaml(fname))
            if ftype == 'ini':
                conf.update(load_ini(fname))

        return conf


def filetype(fname):
    if fname.endswith('.toml'):
        return 'toml'
    elif fname.endswith('.json'):
        return 'json'
    elif fname.endswith('.yaml'):
        return 'yaml'
    elif fname.endswith('.yml'):
        return 'yaml'
    elif fname.endswith('.ini'):
        return 'ini'
    else:
        return 'ini'



# -*- coding: utf-8 -*-
import os
import sys
from tests.filemaker import create_files
from tests.simpledeps import simpledeps
import pytest


def test_relative_imports():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('relimp') == {'relimp.b -> relimp.a'}


def test_relative_imports2():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from . import b
            - b.py: |
                from . import c
            - c.py
    """
    with create_files(files) as workdir:
        deps = simpledeps('relimp')
        assert 'relimp.c -> relimp.b' in deps
        assert 'relimp.b -> relimp.a' in deps


def test_relative_imports3():
    files = """
        relimp:
            - __init__.py
            - a.py: |
                from .b import c
            - b.py
    """
    with create_files(files) as workdir:
        assert simpledeps('relimp') == {'relimp.b -> relimp.a'}


def test_relative_imports_same_name_with_std():
    files = """
        relimp:
            - __init__.py
            - io.py: |
                import io
    """
    with create_files(files) as workdir:
        if sys.version_info < (3,):                # pragma: nocover
            deps = {'relimp.io -> relimp.io'}
        else:                                      # pragma: nocover
            deps = {'io -> relimp.io'}
        assert simpledeps('relimp', '--pylib') == deps


def test_relative_imports_same_name_with_std_future():
    files = """
        relimp:
            - __init__.py
            - io.py: |
                from __future__ import absolute_import
                import io
    """
    with create_files(files) as workdir:
        deps = {
            '__future__ -> relimp.io',
            'io -> relimp.io'
        }
        assert simpledeps('relimp', '--pylib') == deps


def test_pydeps_colors():
    files = """
        pdeps:
            - __init__.py
            - colors.py: |
                import colorsys
            - depgraph.py: |
                import json
                import pprint
                import enum
                from . import colors
    """
    with create_files(files, cleanup=False) as workdir:
        assert simpledeps('pdeps', '-x enum') == {
            'pdeps.colors -> pdeps.depgraph',
        }


def test_hierarchy():
    files = """
        relimp:
            - __init__.py
            - a:
                - __init__.py
                - amodule.py: |
                    from ..b import bmodule
            - b:
                - __init__.py
                - bmodule.py

    """
    with create_files(files, cleanup=True) as workdir:
        os.system("tree " + workdir)
        deps = simpledeps('relimp')
        assert 'relimp.b -> relimp.a.amodule' in deps
        assert 'relimp.b.bmodule -> relimp.a.amodule' in deps
